"""
Copyright 2019 NerdWallet

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import imp
import json
import os
import sys

from terraformpy import compile


def main():
    """Compile *.tf.py files and run Terraform"""
    to_process = [
        ent
        for ent in os.listdir(os.getcwd())
        if ent.endswith('.tf.py')
    ]

    if len(to_process) == 0:
        print("terraformpy - Error loading config: No .tf.py files found in %s" % os.getcwd())
        sys.exit(1)

    print("terraformpy - Processing: %s" % ', '.join(to_process))

    # all we need to do is import our files
    # the nature of resource declaration will register all of the objects for us to compile
    for filename in to_process:
        imp.load_source(filename[:-6], filename)

    # now 'compile' everything that was registered, and write it out the tf.json file
    print("terraformpy - Writing main.tf.json")
    with open('main.tf.json', 'w') as fd:
        json.dump(compile(), fd, indent=4, sort_keys=True)

    if len(sys.argv) > 1:
        print("terraformpy - Running terraform: %s" % ' '.join(sys.argv[1:]))
        # replace ourself with terraform
        os.execvp("terraform", ["terraform"] + sys.argv[1:])



"""
Copyright 2019 NerdWallet

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Terraform "objects"

This module provides a set of classes that can be used to build Terraform configurations in a (mostly) declarative way,
while also leveraging Python to add some functional aspects to automate some of the more repetitive aspects of HCL.
"""
import collections

import six
from schematics.types import compound

from .resource_collections import Variant


def recursive_update(dest, source):
    """Like dict.update, but recursive"""
    for key, val in six.iteritems(source):
        if isinstance(val, collections.Mapping):
            recurse = recursive_update(dest.get(key, {}), val)
            dest[key] = recurse
        else:
            dest[key] = val
    return dest


# Provider names are duplicate keys in the resulting json, so we need a way to represent that
class DuplicateKey(str):
    # Preserving usage order will help plan output remain consistent across invocations
    next_hash_per_str = collections.defaultdict(lambda: 0)

    def __init__(self, key=''):
        super(DuplicateKey, self).__init__(key)

        self.hash = DuplicateKey.next_hash_per_str[key]
        DuplicateKey.next_hash_per_str[key] = self.hash + 1

    def __hash__(self):
        return self.hash


class OrderedDict(compound.DictType):
    """
    A schematic DictType that preserves key insertion order
    """

    def __init__(self, *args, **kwargs):
        super(OrderedDict, self).__init__(*args, **kwargs)
        self.native_type = collections.OrderedDict

    def convert(self, value, context=None):
        # okay, this is a little wasteful for the purpose of not having to
        # copy-paste the implementation that might currently exist
        # in compound.DictType
        # so sure, we iterate over the type a couple of times, but this way
        # we get their validators and updates to their validators
        temp_data = super(OrderedDict, self).convert(value, context)

        # so now, temp_data is a dict filled with the correct (and coerced)
        # values, we just want it to be ordered now based on how `value`
        # was passed in
        # let's trust the validators in `convert()` and not check key existance
        data = collections.OrderedDict()
        for k in value.keys():
            data[k] = temp_data[k]

        return data


class TFObject(object):
    _instances = None
    _frozen = False

    def __new__(cls, *args, **kwargs):
        # create the instance
        inst = super(TFObject, cls).__new__(cls, *args, **kwargs)

        # register it on the class
        try:
            cls._instances.append(inst)
        except AttributeError:
            cls._instances = [inst]

        # return it
        return inst

    @classmethod
    def reset(cls):
        def recursive_reset(cls):
            cls._instances = None
            for klass in cls.__subclasses__():
                recursive_reset(klass)
        recursive_reset(cls)
        TFObject._frozen = False

    @classmethod
    def compile(cls):
        TFObject._frozen = True

        def recursive_compile(cls):
            results = []
            try:
                for instance in cls._instances:
                    output = instance.build()
                    results.append(output)
            except TypeError:
                pass

            for klass in cls.__subclasses__():
                results += recursive_compile(klass)
            return results
        configs = recursive_compile(cls)

        result = {}
        for config in configs:
            result = recursive_update(result, config)
        return result


class Terraform(TFObject):
    """Represents Terraform configuration.

    See: https://www.terraform.io/docs/configuration/terraform.html
    """

    def __init__(self, _values=None, **kwargs):
        self._values = _values or {}
        self._values.update(kwargs)

    def build(self):
        return {
            'terraform': self._values
        }


class NamedObject(TFObject):
    """Named objects are the Terraform definitions that only have a single name component (i.e. variable or output)"""
    # When recursively compiling, the "type" of object that is written out to the terraform definition needs to point to
    # the top-most subclass of TFobject.  For example, if you create a subclass of Resource named MyResource any
    # instances still needs to be written out to the json as "resource" and not "myresource".  This must be set to an
    # appropriate value at each specific subclass that maps to a terraform resource type.
    TF_TYPE = None

    def __init__(self, _name, _values=None, **kwargs):
        """When creating a TF Object you can supply _values if you want to directly influence the values of the object,
        like when you're creating security group rules and need to specify `self`
        """
        assert self.TF_TYPE is not None, "Bad programmer.  Set TF_TYPE on %s" % self.__class__.__name__

        self._name = _name
        self._values = _values or {}

        if Variant.CURRENT_VARIANT is None:
            self._values.update(kwargs)
        else:
            for name in kwargs:
                if not name.endswith('_variant'):
                    self._values[name] = kwargs[name]
                elif name == '{0}_variant'.format(Variant.CURRENT_VARIANT.name):
                    self._values.update(kwargs[name])

    def __setattr__(self, name, value):
        if '_values' in self.__dict__ and name in self.__dict__['_values']:
            self.__dict__['_values'][name] = value
        else:
            self.__dict__[name] = value

    def __getattr__(self, name):
        """This is here as a safety so that you cannot generate hard to debug .tf.json files"""
        if not TFObject._frozen and name in self._values:
            return self._values[name]
        raise AttributeError("%ss does not provide attribute interpolation through attribute access!" %
                             self.__class__.__name__)

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self._name == other._name and self._values == other._values

    def __ne__(self, other):
        return not self.__eq__(other)

    def build(self):
        result = {
            self.TF_TYPE: {
                self._name: self._values
            }
        }
        return result

    def __repr__(self):
        return "{0} {1}".format(type(self), self._name)


class TypedObjectAttr(str):
    """TypedObjectAttr is a wrapper returned by TypedObject for attributes accessed which don't exist.

    The main use case for needing an attr wrapper is accessing interpolated map values, such as those from the
    aws_kms_secrets resource.  We want to support both string-based and integer-based attribute access. Per the
    documentation here:

    https://www.terraform.io/docs/configuration/expressions.html#indices-and-attributes

    However, of note, it appears that several resources - varying by provider version - fail to fully implement both
    specs.  After testing, it appears that dot-separated attributes are more consistently supported. As such, we want
    to return values such as:

    ${resource_type.resource_name.attribute.key_name}
    """

    def __new__(cls, terraform_name, name, item=None):
        obj = super(TypedObjectAttr, cls).__new__(cls, '${{{0}.{1}}}'.format(
            terraform_name,
            cls._name_with_index(name, item)
        ))
        obj._terraform_name = terraform_name
        obj._name = name
        obj._item = item
        return obj

    @staticmethod
    def _name_with_index(name, item):
        if item is None:
            return name
        else:
            return '{0}.{1}'.format(name, item)

    def __getitem__(self, item):
        return TypedObjectAttr(self._terraform_name, self._name_with_index(self._name, self._item), item)

    def __getattr__(self, item):
        return TypedObjectAttr(self._terraform_name, self._name_with_index(self._name, self._item), item)


class TypedObject(NamedObject):
    """Represents a Terraform object that has both a type and name (i.e. resource or data).

    When you access an attribute of an instance of this class it will return the correct interpolation syntax for that
    object.

    .. code-block:: python

        instance = Resource('aws_instance', 'my_instance', ...)
        assert instance.id == '${aws_instance.my_instance.id}'

    This allows you to build up configurations using the instances of the objects as a shortcut to writing out all of
    the interpolation syntax.
    """

    def __init__(self, _type, _name, **kwargs):
        super(TypedObject, self).__init__(_name, **kwargs)
        self._type = _type

        try:
            if Provider.CURRENT_PROVIDER._name == self._type.split('_')[0]:
                self._values['provider'] = Provider.CURRENT_PROVIDER.as_provider()
        except AttributeError:
            # CURRENT_PROVIDER is None
            pass

    def __eq__(self, other):
        return super(TypedObject, self).__eq__(other) and self._type == other._type

    @property
    def terraform_name(self):
        return '.'.join([self._type, self._name])

    def interpolated(self, name):
        """If you reference an attr on a TFObject that was provided as a value you will always get back the python object
        instead of the Terraform interpolation string.  If you want the interpolation string this can be used to get it.
        If you don't do this then you can end up without the implicit dependency and can have failures due to ordering.

        Example:

        .. code-block:: python

            role = Resource(
                'aws_iam_role', 'my_role',
                name='my-role',
                ...
            )
            Resource(
                'aws_iam_role_policy_attachment', 'role_attachment',
                role=role.interpolated('name'),
                ...
            )
        """
        try:
            TFObject._frozen = True
            return getattr(self, name)
        finally:
            TFObject._frozen = False

    def __getattr__(self, name):
        if not TFObject._frozen and name in self._values:
            return self._values[name]
        return TypedObjectAttr(self.terraform_name, name)

    def build(self):
        result = {
            self.TF_TYPE: {
                self._type: {
                    self._name: self._values
                }
            }
        }
        return result

    def __repr__(self):
        return "{0} {1} {2}".format(type(self), self._type, self._name)

    def __str__(self):
        return self.__repr__()


class Provider(NamedObject):
    """Represents a Terraform provider configuration.

    Providers can be used as context managers, and then provide themselves to all objects created while the context is
    active:

    .. code-block:: python

        with Provider("aws", region="us-west-2", alias="west2"):
            sg = Resource('aws_security_group', 'sg', ingress=['foo'])

        assert sg.provider == 'aws.west2'
    """
    TF_TYPE = "provider"
    CURRENT_PROVIDER = None

    def __enter__(self):
        assert self._values['alias'], "Providers must have an alias to be used as a context manager!"
        self._previous_provider = Provider.CURRENT_PROVIDER
        Provider.CURRENT_PROVIDER = self

    def __exit__(self, exc_type, exc_value, traceback):
        Provider.CURRENT_PROVIDER = self._previous_provider

    def as_provider(self):
        return '.'.join([self._name, self._values['alias']])

    # override build to support duplicate key values
    def build(self):
        result = {
            self.TF_TYPE: {
                DuplicateKey(self._name): self._values
            }
        }
        return result


class Variable(NamedObject):
    """Represents a Terraform variable

    You can reference the interpolation syntax for a Variable instance by simply using it as a string.

    .. code-block:: python

        var = Variable('my_var', default='foo')
        assert var == '${var.my_var}'
    """
    TF_TYPE = "variable"

    def __repr__(self):
        return '${{var.{0}}}'.format(self._name)

    def __str__(self):
        return self.__repr__()


class Output(NamedObject):
    """Represents a Terraform output"""
    TF_TYPE = "output"


class Data(TypedObject):
    """Represents a Terraform data source"""
    TF_TYPE = "data"

    @property
    def terraform_name(self):
        return '.'.join(['data', super(Data, self).terraform_name])


class Resource(TypedObject):
    """Represents a Terraform resource"""
    TF_TYPE = "resource"



"""
Copyright 2019 NerdWallet

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import inspect
import os


def relative_file(filename, _caller_depth=1):
    """Given a filename that is relative to the caller of this function this will return a string to be used in
    resource definitions where you need to load a file.

    Given that Terraform's path.module will always point to the main.tf.json path and we need to retrieve files from
    other directories (i.e. from modules) this provides a convenient way to write a file reference without needing
    to do all of the relative pathing and ../../../..

    .. code-block:: python

        class MyThing(ResourceCollection):
            def create_resources(self):
                self.thing = Resource(
                    'some_resource', 'my_thing',
                    template=self.relative_file('files/foo.json')
                )

    And the resulting JSON in main.tf.json will look like:

    .. code-block:: json

        "template": "${file(\"${path.module}/../../../modules/mything/files/foo.json\")}",

    """
    return '${{file("{0}")}}'.format(relative_path(filename, _caller_depth=_caller_depth+1))


def relative_path(path, _caller_depth=1):
    caller = inspect.stack()[_caller_depth]
    return '${{path.module}}/{0}'.format(
        os.path.relpath(os.path.join(os.path.dirname(caller[1]), path))
    )



"""
Copyright 2019 NerdWallet

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import six
from schematics.models import Model

from terraformpy.helpers import relative_file as _relative_file


class ResourceCollection(Model):
    """ResourceCollection is a specialized subclass of the schematics Model object that aims to keep the feel of the
    TFObject while providing full compatibility as a schematics Model.

    Unlike a model where you provide the data as a dict, you provide data as keyword args just like TFObject.

    By default the Variant object is used to lookup variant properites, but you can also provide a variant_name argument
    that will be used instead.

    Variant data is defined inside of a `foo_variant` block.

    .. code-block:: python

        MyResourceColection(
            count=2
            prod_variant=dict(
                count=4
            )
        )

    If the above block was defined within a `Variant('prod')` context then count would be 4, otherwise it would be 2.
    """

    def __init__(self, *args, **kwargs):
        variant_name = kwargs.pop('variant_name', None)

        # if we have positional arguments AND a context then we just want to do the schematics model thing and have
        # super up to the model to let things happen.  this is most likely happening because one resource collection
        # is being used as a reference in a modeltype
        if len(args) > 0 and kwargs.get('context') is not None:
            super(ResourceCollection, self).__init__(*args, **kwargs)
            return

        if variant_name is None and Variant.CURRENT_VARIANT is not None:
            variant_name = Variant.CURRENT_VARIANT.name

            # update our raw data with the variant defaults
            kwargs.update(Variant.CURRENT_VARIANT.defaults)

        if variant_name is not None:
            # if there is then try fetching the val from inside the special variant attr
            variant_key = '{0}_variant'.format(Variant.CURRENT_VARIANT.name)
            variant_data = kwargs.get(variant_key, None)
            if variant_data is not None:
                kwargs.update(variant_data)

        # filter all of the variant data out
        kwargs = dict(
            (k, v)
            for k, v in six.iteritems(kwargs)
            if not k.endswith('_variant')
        )

        super(ResourceCollection, self).__init__(kwargs)

        self.validate()
        self.create_resources()

    def relative_file(self, filename):
        return _relative_file(filename, _caller_depth=2)

    def create_resources(self):
        raise NotImplementedError

    def finalize_resources(self):
        """This is called right before we compile everything.  It gives the collection a chance to generate any final
        resources prior to the compilation occuring.
        """
        pass


class Variant(object):
    """When used as a context manager it provides the ability for ResourceCollection's to vary their inputs based on a
    symbolc string name that allows you to define a resource collection for multiple environments where most of the
    inputs are shared, with only a few differences.

    Any kwargs passed to the constructor become defaults for non-variant inputs.  This allows you to supply inputs that
    are shared between many different ResourceCollections at the variant level so you don't need to pass them over and
    over again.
    """
    CURRENT_VARIANT = None

    def __init__(self, name, **kwargs):
        self.name = name
        self.defaults = kwargs
        self.previous_variant = None

    def __enter__(self):
        self.previous_variant = Variant.CURRENT_VARIANT
        Variant.CURRENT_VARIANT = self

    def __exit__(self, exc_type, exc_value, traceback):
        Variant.CURRENT_VARIANT = self.previous_variant



from setuptools import find_packages, setup

with open('VERSION') as version_fd:
    VERSION = version_fd.read().strip()

with open('README.rst') as readme_fd:
    long_description = readme_fd.read()

setup(
    name='terraformpy',
    version=VERSION,
    author="Evan Borgstrom",
    author_email="eborgstrom@nerdwallet.com",
    description="Terraformpy is a library and command line tool to supercharge "
                "your Terraform configs using a full fledged Python environment!",
    long_description=long_description,
    packages=find_packages('src'),
    package_dir={'': 'src'},
    install_requires=[
        'schematics>=2.0,<3.0',
        'six>=1.11,<2'
    ],
    extras_require={
        'dev:python_version <= "2.7"': [
            'pytest>=4.6,<4.7',
        ],
        'dev:python_version > "3"': [
            'pytest>=5.0,<6',
        ]
    },
    entry_points={
        'console_scripts': [
            'terraformpy = terraformpy.cli:main',
        ],
    },
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Programming Language :: Python",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Pre-processors",
        "Topic :: System :: Systems Administration",
    ]
)
